#! /usr/bin/perl

###########################################################################
# Load configuration from this file:
BEGIN { $CONFFILE = '/var/www/webnews/webnews.conf'; }

# Pull in Perl modules (.pm files) from WebNews installation directory:
BEGIN { $INSTALLDIR = '/var/www/webnews'; }

do $CONFFILE;
use lib $INSTALLDIR;
# ALL OTHER CONFIGURATION CHANGES SHOULD BE MADE IN $CONFFILE, NAMED ABOVE.
###########################################################################

###########################################################################
# You shouldn't have to change these:
# Name of the local copy of the active file, located in $WNDIR.
$ACTIVE = "$WNDIR/active.txt";

# URL by which the active file can be accessed via the web.
$ACTIVEURL = "${BASEURL}active.txt";

# User-Agent header used when posting to newsgroups: 
$USER_AGENT = 'WebNews';

# By default, we trust the username the web server tells us.
$USER = $ENV{'REMOTE_USER'};
###########################################################################

###########################################################################
# Set up the modules we use:
use News::NNTPClient;
use NewsrcLock;
use Net::SMTP;
use Set::IntSpan;
use CGI;
use Text::Wrap;
use HTTP::Date;
use Fcntl qw(:flock);
use MIME::Base64;
use MIME::QuotedPrint;
###########################################################################

###########################################################################
# Squash some warnings.
use vars qw($AUTH_PASSWD $AUTH_USERNAME $MAILHOST $MAINTAINER
            $SERVER_NEEDS_AUTH $SPAMFILTER);
###########################################################################

###########################################################################
# Some CGI preliminaries.
$query = new CGI;
print $query->header;
print $query->start_html(-title=>"WebNews", 
                         -style=>"${BASEURL}/webnews.css");

# Force our errors to be printed into the web page.
$| = 1;
open(STDERR, ">&STDOUT") or die "Couldn't redirect std. err\n";
###########################################################################

###########################################################################
# Set up for the user who called us:
if ($query->param('user')) { $USER = $query->param('user'); }
if (! $USER) {
	print "I can't figure out what your username is. Sorry!\n";
	print signature();
	exit 0;
}
$USERDIR  = "$WNDIR/users/$USER";
$NEWSRC   = "$USERDIR/newsrc";
print "<IMG WIDTH=502 HEIGHT=73 SRC=\"${BASEURL}images/webnews.png\" BORDER=0><BR>\n";
###########################################################################

###########################################################################
# Fix up inputs:
foreach $imageinput ('subscribe', 'Submit', 'chgsubscriptions') {
	if ($query->param("${imageinput}.x") || $query->param("${imageinput}.y")) {
		$query->delete("${imageinput}.x");
		$query->delete("${imageinput}.y");
		$query->param("${imageinput}", 'true');
	}
}
if ($query->param("Submit")) {
	$query->delete("Submit");
	$query->param("submit", "true");
}
###########################################################################

###########################################################################
# Quicken startup time, by not loading newsrc or nntp in certain cases,
# and by not loading active file unless we really need it:
$load_newsrc = 1;
$load_nntp = 1;
$load_active = 0;
if ($query->param('group')) {
	if ($query->param('article')) {
		if ($query->param('submit')) {
			# Submitting an article doesn't require loading the newsrc.
			$load_newsrc = 0;
		}
	} elsif ($query->param('post')) {
		# Just getting the posting form for filling out an
		# entirely new article doesn't require loading the newsrc
		# or connecting to the news server!
		$load_newsrc = 0;
		$load_nntp = 0;
	}
} elsif ($query->param('chgsubscriptions')) {
	# Changing subscriptions needs the active file.
	$load_active = 1;
}
###########################################################################

###########################################################################
# Connect to news server:
umask 0022;
if ($load_nntp) {
	$nntp = new News::NNTPClient($SERVER)
		or die "Couldn't connect to news server $SERVER ($!); " .
			"please try again later.";

	if ($SERVER_NEEDS_AUTH) {
		$nntp->authinfo($AUTH_USERNAME, $AUTH_PASSWD);
		my $auth_errcode = $nntp->code;
		my $auth_errmsg = $nntp->message;
		if ($auth_errcode >= 500) {
			print "<P class=\"wbox\"><B>Error:</B> The NNTP server $SERVER\n";
			print "is refusing authentication ($auth_errcode $auth_errmsg).\n";
			print "Contact your system administrator for more information.</P>";
			print "\n", signature();
			die "Authentication failure ($auth_errcode $auth_errmsg)";
		}
	}

	if ($load_active) {
		# Read in (and, if necessary, update) local copy of news active file:
		if (! -f $ACTIVE) {
			open(ACTIVE, ">$ACTIVE") or die "Can't write active file: $!\n";
			@active = $nntp->list('active');
			1 until flock(ACTIVE, LOCK_EX); # lock the file for write
			print ACTIVE @active;           # write the entire file
			flock(ACTIVE, LOCK_UN);         # unlock the file
			close ACTIVE;
		} else {
			my $since = (stat($ACTIVE))[9]; # get last modification time
			my @newgroups = $nntp->newgroups($since);
			if ($#newgroups != -1) {
				open(ACTIVE, ">>$ACTIVE")
					or die "Can't write active file: $!\n";
				1 until flock(ACTIVE, LOCK_EX); # lock the file for write
				print ACTIVE @newgroups;        # write the new changes
				flock(ACTIVE, LOCK_UN);         # unlock the file
				close ACTIVE;
			}
			open(ACTIVE, "$ACTIVE") or die "Can't read active file: $!\n";
			1 until flock(ACTIVE, LOCK_SH); # lock the file for read
			@active = <ACTIVE>;             # read the entire file
			flock(ACTIVE, LOCK_UN);         # unlock the file
			close ACTIVE;
		}
		@groups = map { s/\s.*$//gs;$_ } @active;
	}

	# We might have timed out waiting for all those stupid locks. Let's ping
	# the server to see whether things are okay. MODE READER is
	# basically a no-op. If we didn't load_active, this should just
	# shoot straight through.
	unless ($nntp->mode_reader) {
		$nntp->quit;
		$nntp->connect;
		if (! $nntp->mode_reader) {
			$err = $nntp->code . " " . $nntp->message;
			$nntp->quit;
			print "<P><B>Error in connecting to news server.</B> [$err]</P>\n";
			print signature();
			exit 0;
		}
	}
}
###########################################################################

###########################################################################
# Load the newsrc:
if ($load_newsrc) {
	# We may have a new user here. Make a new userdir and/or newsrc if so.
	if (! -d $USERDIR) {
		mkdir $USERDIR, 0777;
	}
	$newsrc = new NewsrcLock;
	if (! -f $NEWSRC) {
		$newsrc->subscribe('news.announce.newusers');
		$newsrc->save_as($NEWSRC);
	}
	$newsrc->load($NEWSRC) or die "Couldn't load $NEWSRC: $!";
	# The following flag should be set to nonzero on any newsrc
	# object modification:
	$newsrc_modified = 0;
}
###########################################################################

###########################################################################
# Useful for debugging:
#print "<UL>\n";
#foreach $name ($query->param) {
#	$value = $query->param($name);
#	print "<LI>$name = \"$value\"</LI>\n";
#}
#print "</UL>\n";
###########################################################################

###########################################################################
# Main dispatch tree:
if ($query->param('group')) {
	if ($query->param('unsubscribe')) {
		do_unsubscribe($query->param('group'));
	} elsif ($query->param('catchup')) {
		do_catchup($query->param('group'));
	} elsif ($query->param('post')) {
		do_post($query->param('group'));
	} elsif ($query->param('article')) {
		if ($query->param('submit')) {
			do_submit($query->param('group'), $query->param('article'));
		} elsif ($query->param('reply')) {
			do_reply($query->param('group'), $query->param('article'));
		} else {
			do_article($query->param('group'), $query->param('article'));
		}
	} else {
		do_group($query->param('group'));
	}
} elsif ($query->param('chgsubscriptions')) {
	do_chgsubscriptions();
} else {
	do_groupindex();
}
###########################################################################

$unsubscribed_group = '';

sub unqp {
	my($op, $s) = @_;
	if (lc $op eq 'q') {
		$s =~ tr/_/ /;
		$s = MIME::QuotedPrint::decode_qp $s;
	} else {
		$s = MIME::Base64::decode_base64 $s;
	}
	$s;
}

sub unrfc {
	for (@_) {
		s/=\?(?:iso-8859-15?|windows-1252)\?([bq])\?(.+?)\?=/unqp($1, $2)/iges;
	}
}

sub start_headers {
	return "<TABLE BORDER=0>\n";
}

sub end_headers {
	return "</TABLE><P>\n";
}

sub print_header {
	my ($full, $field, $contents) = @_;
	unrfc $contents;
	return if not $full and $field !~ /^(From|Newsgroups|Subject|Date|Organization|Distribution)/io;
	$contents = $query->escapeHTML($contents);
	print "<TR><TD ALIGN=RIGHT VALIGN=TOP><B>${field}</B></TD>",
		"<TD VALIGN=TOP>${contents}</TD></TR>\n";
}

sub fetch_article {
	my ($nntp, $group, $article) = @_;
	$nntp->group($group);
	unless ($newsrc->marked($group, $article)) {
		$newsrc->mark($group, $article);
		$newsrc_modified++;
	}
	return $nntp->article($article);
}

sub artindex {
	my ($group) = @_;
	$query->delete('article');
	$query->param('group', $group);
	my $selfurl = $query->self_url;
	return "<A HREF=\"$selfurl\" class=\"btn\">Article index</A> \n";
}

sub grpindex {
	$query->delete('article');
	$query->delete('group');
	my $selfurl = $query->self_url;
	return "<A HREF=\"$selfurl\" class=\"btn\">Group index</A>\n";
}

sub article_n {
	my($next, $text) = @_;
	if (not $next) { return "<div class=\"gbtn\">No $text</div>\n" }
	my $cur = $query->param('article');
	$query->param('article', $next);
	my $selfurl = $query->self_url;
	$query->param('article', $cur);
	return "<A HREF=\"$selfurl\" class=\"btn\">$text</A>\n";
}

sub toggleunread {
	$query->delete('article');
	$query->delete('group');
	if ($_[0]) {
		$query->param('group', $_[0]);
	}
	if ($query->param('showold')) {
		$query->delete('showold');
	} else {
		$query->param('showold', 'true');
	}
	my $selfurl = $query->self_url;
	if ($query->param('showold')) {
		$query->delete('showold');
	} else {
		$query->param('showold', 'true');
	}
	$query->delete('group');
	return "<A HREF=\"$selfurl\" class=\"btn\">Toggle unread</a>\n";
}

sub unsubscribe {
	my ($group) = @_;
	$query->param('group', $group);
	$query->param('unsubscribe', 'true');
	my $selfurl = $query->self_url;
	$query->delete('unsubscribe');
	return "<A HREF=\"$selfurl\" class=\"btn\">Unsubscribe</A>\n";
}

sub showhdrs {
	my ($group, $article, $full) = @_;
	my ($oldhdrs);
	my $full_brief = $full ? 'brief' : 'full';
	$query->param('group', $group);
	$query->param('article', $article);
	$oldhdrs = $query->param('headers');
	$query->param('headers', $full_brief);
	my $selfurl = $query->self_url;
	$query->param('headers', $oldhdrs);
	return "<A HREF=\"$selfurl\" class=\"btn\">\u$full_brief headers</A>\n";
}

sub reply {
	my ($group, $article) = @_;
	$query->param('group', $group);
	$query->param('article', $article);
	$query->param('reply', 'true');
	my $selfurl = $query->self_url;
	$query->delete('reply');
	return "<A HREF=\"$selfurl\" class=\"btn\">Follow-up</A>\n";
}

sub returntoarticle {
	my ($group, $article) = @_;
	$query->param('group', $group);
	$query->param('article', $article);
	$query->delete('reply');
	my $selfurl = $query->self_url;
	return "<A HREF=\"$selfurl\" class=\"btn\">Return to article</A>\n";
}

sub post {
	my ($group) = @_;
	$query->param('group', $group);
	$query->param('post', 'true');
	my $selfurl = $query->self_url;
	$query->delete('post');
	return "<A HREF=\"$selfurl\" class=\"btn\">Post message</A>\n";
}

sub catchup {
	my ($group) = @_;
	$query->param('group', $group);
	$query->param('catchup', 'true');
	my $selfurl = $query->self_url;
	$query->delete('catchup');
	return "<A HREF=\"$selfurl\" class=\"btn\">Catch up</A>\n";
}

sub sortbynumber {
	my ($group) = @_;
	$query->param('group', $group);
	$query->delete('sortorder');
	$query->param('sortorder', 'number');
	my $selfurl = $query->self_url;
	return "<A HREF=\"$selfurl\" class=\"sortby\">Number</A>\n";
}

sub sortbyfrom {
	my ($group) = @_;
	$query->param('group', $group);
	$query->delete('sortorder');
	$query->param('sortorder', 'from');
	my $selfurl = $query->self_url;
	return "<A HREF=\"$selfurl\" class=\"sortby\">Author</A>\n";
}

sub sortbysubject {
	my ($group) = @_;
	$query->param('group', $group);
	$query->delete('sortorder');
	$query->param('sortorder', 'subject');
	my $selfurl = $query->self_url;
	return "<A HREF=\"$selfurl\" class=\"sortby\">Subject</A>\n";
}

sub sortbydate {
	my ($group) = @_;
	$query->param('group', $group);
	$query->delete('sortorder');
	$query->param('sortorder', 'date');
	my $selfurl = $query->self_url;
	return "<A HREF=\"$selfurl\" class=\"sortby\">Date</A>\n";
}

sub prettyfrom {
  my $s = shift;
  $s =~ s/^\s*(\S.*\S)\s*<.+?\@.*?\w>\s*$/$1/;   # Hans Foo <hans@xy>
  $s =~ s/^\s*\S+\@\S+\s+\(\s*(\S.*\S)\s*\)/$1/; # hans@xyz (Hans Foo)
  return $query->escapeHTML($s);
}

sub gecos_name {
	my ($user) = @_;
	my ($name,$passwd,$uid,$gid,$quota,$comment,$gcos,$dir,$shell,$expire)
		= getpwnam($user);
	return $user unless $gcos;
	$gcos =~ s/&/ucfirst $name/e if $gcos =~ /&/;
	$gcos =~ s/,.*$// if $gcos =~ /,/;
	return $gcos;
}

sub wrap_line {
	$Text::Wrap::columns = 79;
	return split ("\n", wrap('','', $_[0]));
}

sub wrap_message {
	my @input = @_;
	my @output;
	foreach my $line (@input) {
		if ($line !~ /^\s*>/ and length($line) >= 79) { # don't wrap quoted lines
			push(@output, wrap_line($line));
		} else {
			push(@output, $line);
		}
	}
	return @output;
}

sub do_submit {
	my ($group, $article) = @_;
	my ($fromaddr, $from, $newsgroups, $subject, $followup_to, $summary,
		$organization, $distro, $replyto, $cc, $references);
	my ($wordwrap, $article_body, $msgid);
	$query->delete('submit');
	my $full  = gecos_name($USER);
	$fromaddr = "$USER\@$DOMAIN";
	$fromaddr = "spamfilter\@$DOMAIN" if $SPAMFILTER;
	$from = "$full <$fromaddr>";
	$wordwrap     = $query->param('wordwrap');     $query->delete('wordwrap');
	$newsgroups   = $query->param('newsgroups');   $query->delete('newsgroups');
	$subject      = $query->param('subject');      $query->delete('subject');
	$followup_to  = $query->param('followup_to');  $query->delete('followup_to');
	$summary      = $query->param('summary');      $query->delete('summary');
	$organization = $query->param('organization'); $query->delete('organization');
	$distro       = $query->param('distro');       $query->delete('distro');
	$replyto      = $query->param('replyto');      $query->delete('replyto');
	$cc           = $query->param('cc');           $query->delete('cc');
	$article_body = $query->param('article_body'); $query->delete('article_body');
	$msgid        = $query->param('msgid');        $query->delete('msgid');
	$references   = $query->param('references');   $query->delete('references');
	print $query->h2("[$USER] Submitting new article to $group"), "\n";
	print grpindex();
	print artindex($group);
	print unsubscribe($group);
	print returntoarticle($group, $article) if ($article != 0);
	if (! $nntp->postok) {
		print "<P class=\"wbox\"><B>Error:</B> The NNTP server $SERVER is not allowing \n";
		print "posting. Contact your system administrator for \n";
		print "more information.</P>\n";
		print signature();
		return;
	}
	if (! $subject) {
		print "<P class=\"wbox\"><B>Error:</B> You failed to give a subject. Go 'back' in \n";
		print "your browser and fill one in, if you want to post your\n";
		print "message.</P>\n";
		print signature();
		return;
	}
	push(@headers, "From: $from\n");
	push(@headers, "Newsgroups: $newsgroups\n");
	push(@headers, "Subject: $subject\n") if $subject;
	push(@headers, "Followup-To: $followup_to\n") if $followup_to;
	push(@headers, "References: $references\n") if $references;
	push(@headers, "Summary: $summary\n") if $summary;
	push(@headers, "Organization: $organization\n") if $organization;
	push(@headers, "Distribution: $distro\n") if $distro;
	push(@headers, "Reply-To: $replyto\n") if $replyto;
	push(@headers, "Cc: $cc\n") if $cc;
	push(@headers, "User-Agent: $USER_AGENT\n");
	$article_body =~ s/\r//g;
	my @article_body = split(/\n/, $article_body);
	@article_body = wrap_message(@article_body) if $wordwrap eq 'true';
	$result = $nntp->post(@headers, "\n", @article_body);
	if ($result) {
		$result = $nntp->code . " " . $nntp->message;
		chomp $result;
		print "<P class=\"wbox\">Your message was posted successfully.<br>[$result]</P>\n";
	} else {
		$result = $nntp->code . " " . $nntp->message;
		chomp $result;
		print "<P class=\"wbox\"><B>Error in posting.</B> Your message was NOT sent.<br>Server response: [$result]</P>\n";
	}
	if ($cc) {
		$smtp = Net::SMTP->new($MAILHOST);
		$smtp->mail($fromaddr);
		$smtp->to($cc);
		$smtp->data();
		$smtp->datasend(@headers, "\n", join("\n", @article_body));
		$smtp->dataend();
		$smtp->quit;
	}
	print signature();
}

sub do_reply {
	my ($group, $article) = @_;
	my ($msgid, $from, $newsgroups, $subject);
	my $references = '';

	my @article = fetch_article($nntp, $group, $article);

	my @hdr = ();
	while (@article) {
		my $l = shift @article;
		$l =~ s/\s+$//;
		last if $l eq ''; 
		if (@hdr and $l =~ /^\s/) { $hdr[-1] .= $l; } else { push @hdr, $l; }
	}

	for my $line (@hdr) {
		next unless ($field, $contents) = $line =~ /^([^:]*:) (.*)$/;
		if ($field =~ /Message-ID/io) {
			$msgid = $contents;
		} elsif ($field =~ /From/io) {
			$from = $contents;
		} elsif ($field =~ /Newsgroups/io) {
			$newsgroups = $contents;
		} elsif ($field =~ /Subject/io) {
			($subject = $contents) =~ s/^(Re:\s*)*/Re: /;
		} elsif ($field =~ /References/io) {
			$references = $contents;
		}
	}
	$references .= " " . $msgid;
	@article = map { "> $_"; } @article;
	# Simple attribution ``* Andi Karrer <xx@yy.ch>:'', recommended by GNKSA
	my $article_body = join "", "* $from:\n", @article;
	$query->delete('reply');
	$query->delete('user');
	$query->delete('group');
	my $selfurl = $query->self_url;
	
	print $query->h2("[$USER] Follow-up to article $article in $group"), "\n";
	print "<FORM METHOD=\"POST\" ACTION=\"$selfurl\">\n";
	return do_postingform($references, $group, $article, $USER, $msgid,
		$newsgroups, $subject, $DEFAULT_ORG, $DEFAULT_DISTRO,
		$article_body);
}

sub do_post {
	my ($group) = @_;
	$query->delete('post');
	$query->delete('user');
	$query->delete('group');
	my $selfurl = $query->self_url;

	print $query->h2("[$USER] Post new message to $group"), "\n";
	print "<FORM METHOD=\"POST\" ACTION=\"$selfurl\">\n";
	return do_postingform("", $group, "0 but true", $USER, "",
		$group, "", $DEFAULT_ORG, $DEFAULT_DISTRO, "");
}

sub do_postingform {
	my ($follow_up, $group, $article, $user, $msgid,
		$newsgroups, $subject, $org, $distro,
		$article_body) = @_;
	$query->param('user', $user);
	$query->param('group', $group);
	print grpindex();
	print artindex($group);
	print unsubscribe($group);
	print "<INPUT TYPE=\"hidden\" NAME=\"article\" VALUE=\"$article\">\n";
	if ($follow_up) {
		print returntoarticle($group, $article);
		print "<INPUT TYPE=\"hidden\" NAME=\"msgid\" VALUE=\"$msgid\">\n";
		print "<INPUT TYPE=\"hidden\" NAME=\"references\" VALUE=\"$follow_up\">\n";
	}

	print "<INPUT TYPE=\"hidden\" NAME=\"group\" VALUE=\"$group\">\n";
	print "<INPUT TYPE=\"hidden\" NAME=\"user\" VALUE=\"$USER\">\n";
	print "<INPUT TYPE=\"hidden\" NAME=\"Submit\" VALUE=\"false\">\n";
	print "<a href=\"javascript:document.forms[0].Submit.value='true';document.forms[0].submit();\" class=\"btn\">Submit</a>\n";


	print "<P>(If you wish to cancel posting, ";
	print "click <B>Return to Article</B> to go back to what you were" .
		" reading, " if $follow_up;
	print "click <B>Article Index</B> to read more articles from this group, ";
	print "or click <B>Group Index</B> to choose a different group.)</P>\n";

	print "<TABLE>\n";
	print "<TR><TD ALIGN=RIGHT>Newsgroups:</TD><TD ALIGN=LEFT>", 
	"<INPUT TYPE=\"text\" SIZE=70 NAME=\"newsgroups\" VALUE=\"$newsgroups\">",
	"</TD></TR>\n";
	print "<TR><TD ALIGN=RIGHT>Subject:</TD><TD ALIGN=LEFT>", 
	"<INPUT TYPE=\"text\" SIZE=70 NAME=\"subject\" VALUE=\"$subject\">",
	"</TD></TR>\n";
	print "<TR><TD ALIGN=RIGHT>Followup-To:</TD><TD ALIGN=LEFT>", 
	"<INPUT TYPE=\"text\" SIZE=70 NAME=\"followup_to\" VALUE=\"$newsgroups\">",
	"</TD></TR>\n";
	print "<TR><TD ALIGN=RIGHT>Summary:</TD><TD ALIGN=LEFT>", 
	"<INPUT TYPE=\"text\" SIZE=70 NAME=\"summary\">",
	"</TD></TR>\n";
	print "<TR><TD ALIGN=RIGHT>Organization:</TD><TD ALIGN=LEFT>", 
	"<INPUT TYPE=\"text\" SIZE=70 NAME=\"organization\" VALUE=\"$DEFAULT_ORG\">",
	"</TD></TR>\n";
	print "<TR><TD ALIGN=RIGHT>Distribution:</TD><TD ALIGN=LEFT>", 
	"<INPUT TYPE=\"text\" SIZE=70 NAME=\"distro\" VALUE=\"$DEFAULT_DISTRO\">",
	"</TD></TR>\n";
	print "<TR><TD ALIGN=RIGHT>Reply-To:</TD><TD ALIGN=LEFT>", 
	"<INPUT TYPE=\"text\" SIZE=70 NAME=\"replyto\">",
	"</TD></TR>\n";
	print "<TR><TD ALIGN=RIGHT>Cc:</TD><TD ALIGN=LEFT>", 
	"<INPUT TYPE=\"text\" SIZE=70 NAME=\"cc\">",
	"</TD></TR>\n";
	print "</TABLE>\n";
	print "<HR>\n";
	print $query->checkbox(-name=>'wordwrap', -checked=>'yes',
		-value=>'true', -label=>'Auto-fill to 80 columns');
	print "<BR>\n<TEXTAREA COLS=80 ROWS=24 NAME=\"article_body\">\n";
	$article_body = $query->escapeHTML($article_body);
	print "$article_body\n";
	print "</TEXTAREA>";
	print "</FORM><P>";
}

sub do_catchup {
	my ($group) = @_;
	my $got_params = ($query->param('catchup_from')) ? 1 : 0;
	my $catchup_from = $query->param('catchup_from');
	my $catchup_to = $query->param('catchup_to');
	$query->delete('submit');
	$query->delete('catchup');
	$query->delete('catchup_from');
	$query->delete('catchup_to');
	my $selfurl = $query->self_url;

	my ($first, $last) = $nntp->group($group);
	if ($got_params) {
		$newsrc->mark_range($group, $catchup_from, $catchup_to);
		$catchup_from--; $catchup_to++;
		if ($catchup_from > $first) {
			$newsrc->unmark_range($group, $first, $catchup_from)
		}
		if ($catchup_to   < $last ) {
			$newsrc->unmark_range($group, $catchup_to, $last)
		}
		$newsrc_modified++;
		return do_group($group);
	}

	print "<FORM METHOD=POST ACTION=\"$selfurl\">\n";
	print "<INPUT TYPE=\"hidden\" NAME=\"group\" VALUE=\"$group\">\n";
	print "<INPUT TYPE=\"hidden\" NAME=\"user\" VALUE=\"$USER\">\n";
	print "<INPUT TYPE=\"hidden\" NAME=\"catchup\" VALUE=\"true\">\n";
	print grpindex();
	print artindex($group);
	print "<INPUT TYPE=\"hidden\" NAME=\"Submit\" VALUE=\"false\">\n";
	print "<a href=\"javascript:document.forms[0].Submit.value='true';document.forms[0].submit();\" class=\"btn\">Catch up</a>\n";

	print $query->h2("[$USER] Catch up articles in $group"), "\n";
	
	print "<P>Click <B>Catch up</B>, above, if you would like to mark the range of articles below \n";
	print "as read.<br>";
	print "Just hit <B>Article Index</B> if you want to cancel.</P>\n";

	print "<P class=\"wbox\">Mark articles from: \n";
	print $query->textfield(-name=>'catchup_from', -default=>"$first",
		-size=>7);
	print " to \n";
	print $query->textfield(-name=>'catchup_to', -default=>"$last",
		-size=>7);
	print " as read (and all others as unread).</P>\n";

	print "</FORM>\n";
}

sub do_article {
	my ($group, $article) = @_;

	my $readarticles = Set::IntSpan->new(scalar $newsrc->marked_articles($group));
	my $serverhas = Set::IntSpan->new(scalar $nntp->group($group));
	my $newarticles = $serverhas->diff($readarticles);
	my $nxunread = '';
	for ($newarticles->elements) {
		if ($_ > $article) { $nxunread = $_; last; }
	}
	my $prev = ''; my $next = '';
	for ($serverhas->elements) {
		if ($_ < $article) { $prev = $_; }
		if ($_ > $article) { $next = $_; last; }
	}

	my @article = fetch_article($nntp, $group, $article);
	my $i = 0;
	my $selfurl = $query->url(-absolute);

	print $query->h2("[$USER] Article number $article in $group"), "\n";
	my $full = (($query->param('headers') eq 'full') ? 1 : 0);
	$query->delete('headers');
	print article_n($nxunread, 'Next unread');
	print article_n($next, 'Next article');
	print article_n($prev, 'Prev article');
	print grpindex();
	print artindex($group);
	print unsubscribe($group);
	print reply($group, $article);
	print showhdrs($group,$article, $full);
	print "<p/>";
	print start_headers();

	my @hdr = ();
	while (@article) {
		my $l = shift @article;
		$l =~ s/\s+$//;
		last if $l eq '';
		if (@hdr and $l =~ /^\s/) { $hdr[-1] .= $l} else { push @hdr, $l }
	}
	for my $line (@hdr) {
		if ($line =~ /^([^:]*:)( .*)$/) {
			print_header($full, $1, $2);
		} else { 
			print_header(1,"(bad line in header):", $line);
		}
	}

	print end_headers();
	print "\n<PRE class=\"wbox\">";
	for (@article) {
		my $c = 0;
		if (/^((\s*>)+)/) {
			$c = $1;
			$c =~ s/\s+//g;
			$c = length($c) % 3 + 1;
			print "<div class=\"quote$c\">";
		}
		print $query->escapeHTML($_);
		print "</div>" if $c;
	}
	print "</PRE><P>\n";
}

sub xover_sort_by_subject {
	local ($anumb,$asubj,$afrom,$adate,$amesg,$arefr,$achar,$aline,$axref) 
		= split(/\t/, $a);
	local ($bnumb,$bsubj,$bfrom,$bdate,$bmesg,$brefr,$bchar,$bline,$bxref) 
		= split(/\t/, $b);
	1 while $asubj =~ s/^Re(:|-)?\s*//g;
	1 while $bsubj =~ s/^Re(:|-)?\s*//g;
	local ($sortorder) = (uc($asubj) cmp uc($bsubj));
	return $sortorder ? $sortorder : ($anumb <=> $bnumb);
}

sub xover_sort_by_from {
	local ($anumb,$asubj,$afrom,$adate,$amesg,$arefr,$achar,$aline,$axref) 
		= split(/\t/, $a);
	local ($bnumb,$bsubj,$bfrom,$bdate,$bmesg,$brefr,$bchar,$bline,$bxref) 
		= split(/\t/, $b);
	$afrom =~ s/\"//g;
	$bfrom =~ s/\"//g;
	local ($sortorder) = (uc($afrom) cmp uc($bfrom));
	return $sortorder ? $sortorder : ($anumb <=> $bnumb);
}

sub xover_sort_by_date {
	local ($anumb,$asubj,$afrom,$adate,$amesg,$arefr,$achar,$aline,$axref) 
		= split(/\t/, $a);
	local ($bnumb,$bsubj,$bfrom,$bdate,$bmesg,$brefr,$bchar,$bline,$bxref) 
		= split(/\t/, $b);
	local ($sortorder) = (str2time($adate) <=> str2time($bdate));
	return $sortorder ? $sortorder : ($anumb <=> $bnumb);
}

sub xover_sort_by_number {
	local ($anumb,$asubj,$afrom,$adate,$amesg,$arefr,$achar,$aline,$axref) 
		= split(/\t/, $a);
	local ($bnumb,$bsubj,$bfrom,$bdate,$bmesg,$brefr,$bchar,$bline,$bxref) 
		= split(/\t/, $b);
	return ($anumb <=> $bnumb);
}

sub xover_sort {
	my ($sort_order, @xover) = @_;
	$sort_order = "xover_sort_by_$sort_order";
	return sort $sort_order @xover;
}

sub do_group {
	my ($group) = @_;
	my ($sortorder) = 'number';
	my (@rows);
	my ($readarticles, $serverhas, $newarticles);

	print $query->h2("[$USER] Article index for $group"), "\n";
	print grpindex();
	print unsubscribe($group);
	print post($group);
	print catchup($group);
	print toggleunread($group);

	if ($query->param('sortorder')) {
		$sortorder = $query->param('sortorder');
	}

	$readarticles = Set::IntSpan->new(scalar $newsrc->marked_articles($group));
	$serverhas = Set::IntSpan->new(scalar $nntp->group($group));
	$newarticles = $serverhas->diff($readarticles);

	# Fields in xover: numb subj from date mesg refr char line xref
	@xover = $nntp->xover($serverhas->run_list);
	unrfc @xover;

	if ($query->param('showold') eq 'true') {
		print "<p>Showing <b>all</b> articles.<br>Click <b>Toggle Unread</b> to hide the articles you've already read.<br>";
	} else {
		print "<p>Showing only <b>new</b> articles.<br>Click <b>Toggle Unread</b> to show all available articles.<br>";
	}

	print "When you're done with this group, please click on <B>Group Index</B> to select another.</p>\n";

	print "<div class=\"wbox\"><TABLE>";

	print $query->Tr($query->td([sortbynumber($group), sortbyfrom($group), sortbysubject($group), sortbydate($group)])), "\n";
	foreach my $hdr (xover_sort($sortorder, @xover)) {
		my (@hdr) = split(/\t/, $hdr);
		my ($numb,$subj,$from,$date,$mesg,$refr,$char,$line,$xref) = @hdr;
		next unless ($newarticles->member($numb) || ($query->param('showold') eq 'true'));
		$query->param('group',$group);
		$query->param('article',$numb);
		my $selfurl = $query->self_url;
		$numb = "<A HREF=\"$selfurl\">"
			. ($newarticles->member($numb) ? "<B>" : "")
			. $numb
			. ($newarticles->member($numb) ? "</B>" : "")
			. "</A>";
		$from = prettyfrom($from);
		$subj = $query->escapeHTML($subj);
		$date = time2str(str2time($date));
		print $query->Tr($query->td([$numb,$from,$subj,$date])), "\n";
	}
	print "</TABLE></div><P>\n";
}

sub signature {
	return "<ADDRESS>${MAINTAINER}</ADDRESS>\n";
}

sub do_unsubscribe {
	my ($group) = @_;
	$unsubscribed_group = $group;
	$newsrc->unsubscribe($group);
	$newsrc_modified++;
	$query->delete('unsubscribe');
	do_groupindex();
}

sub chgsubscriptions {
	$query->param('chgsubscriptions', 'true');
	$query->delete('group');
	my $selfurl = $query->self_url;
	$query->delete('chgsubscriptions');
	return "<A HREF=\"$selfurl\" class=\"btn\">Change Subscriptions</a>\n";
}

sub do_chgsubscriptions {
	my $subs = $query->param('subs');
	my @unsubs = $query->param('unsubs');
	my $selfurl = $query->self_url;
	$query->delete('subs');
	$query->delete('unsubs');
	$query->delete('chgsubscriptions');
	print $query->h2("[$USER] Change subscriptions"), "\n";
	print "<FORM METHOD=\"POST\" ACTION=\"$selfurl\">\n";
	print grpindex();
	print "<INPUT TYPE=\"hidden\" NAME=\"chgsubscriptions\" VALUE=\"true\">\n";
	print "<INPUT TYPE=\"hidden\" NAME=\"user\" VALUE=\"$USER\">\n";
	print "<a href=\"javascript:document.forms[0].chgsubscriptions.value='true';document.forms[0].submit();\" class=\"btn\">Submit</a>\n";
	# Anytime we get here, the newsrc will need to be rewritten.
	$newsrc_modified++;
	if ($subs || @unsubs) {
		print "<BR><H3>THE FOLLOWING UPDATES TO YOUR SUBSCRIPTIONS HAVE BEEN MADE:</H3><P>";
	}
	if ($subs) {
		my @subs = split(/\n/, $subs);
		foreach my $g (@subs) {
			$g =~ s/^\s*//; $g =~ s/\s*$//;
			if ($g =~ /^([^:!]*)([:!]) ([0-9,\-]*)$/) {
				my ($ng, $su, $al) = ($1, $2, $3);
				if (! grep {$_ eq $ng;} @groups) {
					print "<B>Can't use group $ng because the server doesn't have it.</B><BR>\n";
				} else {
					if (! $newsrc->exists($ng)) {
						$newsrc->add_group($ng); 
					}
					if ($su eq ":") {
						$newsrc->subscribe($ng);
					} elsif ($su eq "!") {
						$newsrc->unsubscribe($ng);
					}
					$newsrc->set_articles($ng, $al);
					print "<B>Imported <CODE>.newsrc</CODE> line: $g</B><BR>\n";
				}
			} else {
				if (! grep {$_ eq $g;} @groups) {
					print "<B>Can't find $g in the list of active newsgroups</B><BR>\n";
				} else {
					if (! $newsrc->exists($g)) {
						if (! $newsrc->add_group($g)) {
							print "<B>Failed to add $g to your newsgroup list.</B><BR>\n";
							# newsgroups are added in a subscribed state,
							# so we are done.
						} else {
							print "<B>You have been subscribed to $g.</B><BR>\n";
						}
					} elsif ($newsrc->subscribed($g)) {
						print "<B>Can't subscribe to a group you are already subscribed to [$g].</B><BR>\n";
					} elsif ($newsrc->subscribe($g)) {
						print "<B>You have been subscribed to $g.</B><BR>\n";
					} else {
						print "<B>Failed to subscribe to $g.</B><BR>\n";
					}
				}
			}
		}
	}
	if (@unsubs) {
		foreach my $g (@unsubs) {
			if (! $g) {
				print "<B>Can't unsubscribe from an empty group name.</B><BR>\n";
			} elsif (! $newsrc->subscribed($g)) {
				print "<B>Can't unsubscribe from a group you are not already subscribed to [$g].</B><BR>\n";
			} else {
				$newsrc->unsubscribe($g);
			}
			unless ($newsrc->subscribed($g)) {
				print "<B>You have been unsubscribed from $g.</B><BR>\n";
			} else {
				print "<B>Failed to unsubscribe from $g.</B><BR>\n";
			}
		}
	}
	print "<p>(When you're done here, click <b>Group Index</b> to go ";
	print "back to the list of newsgroups you've subscribed to.)</p>";
	print "<H3>SUBSCRIBING (<A HREF=\"#unsub\">see below for unsubscribing</A>)</H3>\n";

	my $sizekb = 1 + int((stat $ACTIVE)[7] / 1024);
	if ($sizekb > 5) {
		print "<P>You may browse the list of active newsgroups ";
		print "<A HREF=\"${ACTIVEURL}\">here</A>, but be warned, ";
		print "it's very long (currently $sizekb kilobytes).</P>\n";
	} else {
		open A, $ACTIVE;
		my @active = map { s/\s.*$/\n/s; $_ } <A>;
		close A;
		print "<p>Available newsgroups:\n<pre>", @active, "</pre>\n";
	}

	print "<P>You can subscribe to groups by typing their names\n";
	print "in here. Specify one newsgroup on each line, and click\n";
	print "<B>Submit</B> when you're done.\n";
	print "Alternatively, you may paste your <CODE>.newsrc</CODE>\n";
	print "in this box. (Note: You probably don't want to subscribe to</P>\n";
	print "anything named <CODE>junk</CODE> or <CODE>control</CODE>.)</P>\n";
	print "<TEXTAREA COLS=80 ROWS=12 NAME=\"subs\">\n";
	print "</TEXTAREA>";
	print "<A NAME=\"unsub\"><H3>UNSUBSCRIBING</H3></A>\n";
	print "<P>Click the checkbox next to the group name to unsubscribe ";
	print "from that group. When you have checked all the groups you ";
	print "wish to unsubscribe from, click <B>Submit</B>, above. </P>\n";
	my @groups = $newsrc->sub_groups;
	my %labels;
	foreach (@groups) { $labels{$_} = "Unsubscribe from $_"; }
	print $query->checkbox_group(-name=>'unsubs', '-values'=>\@groups,
		-labels=>\%labels, -linebreak=>'true');
}

sub do_groupindex {
	print $query->h2("[$USER] Group index"), "\n";
	print chgsubscriptions();
	print toggleunread();
	print "<p>";
	if ($unsubscribed_group) {
		print "<B>You have been unsubscribed from ${unsubscribed_group}.</B><BR>\n";
	}
	if ($query->param('showold') eq 'true') {
		print "WebNews will show you <b>all</b> the articles. Click <b>Toggle Unread</b> to hide articles you've already read.</p>";
	} else {
		print "WebNews will only show you <b>new</b> articles. Click <b>Toggle Unread</b> if you want to see all the articles.</p>";
	}
	my @sub_groups = $newsrc->sub_groups;
	if ($#sub_groups >= 0) {
		print "\n<P>Your ", $#sub_groups + 1, " current subscription",
			($#sub_groups ? "s" : "") , ":</P>\n";
		print "<P class=\"wbox\">\n";
		foreach my $group (@sub_groups) {
			my ($readarticles, $serverhas, $newarticles);
			$query->param('group',$group);
			$selfurl = $query->self_url;
			print "<A HREF=\"$selfurl\">$group</A>: ";
			$readarticles = Set::IntSpan->new(scalar $newsrc->marked_articles($group));
			$serverhas = scalar $nntp->group($group);
			if (! $serverhas) {
				$newsrc->del_group($group);
				$newsrc_modified++;
				print "<BR><B>Removed group $group from your subscriptions, because the server doesn't have it.</B><BR>\n";
			}
			$serverhas = Set::IntSpan->new($serverhas);
			$howmany = $serverhas->cardinality;
			print $serverhas->cardinality, " article",
				(($howmany > 1) ? "s" : "");
			$newarticles = $serverhas->diff($readarticles);
			print ", <B>", $newarticles->cardinality, " unread</B>"
				if (! $newarticles->empty);
			print ".<BR>\n";
		}
		print "</P>\n";
	} else {
		print "<P class=\"wbox\"><B>You are not currently subscribed to any groups.</B>\n";
		print "If you wish to subscribe to some newsgroups, click on\n";
		print "<B>Change Subscriptions</B>, above.</P>\n";
	}
	print signature();
}

$newsrc->save if $load_newsrc and $newsrc_modified;
$nntp->quit   if $load_nntp;

print $query->end_html;

exit 0;
